<?php

/*
 * Copyright (c) 2012, Steve "Sc00bz" Thomas <steve AT tobtu DOT com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * This is for encrypted bcrypt hashes.
 * The idea behind encrypting the hashes is to protect against a SQL injection only attack.
 *
 * *** DO NOT STORE THE KEY IN THE DATABASE ***
 *
 * WARNING: bcrypt has a max password length of 72 bytes so if you want to get rid of that
 * limit then you should SHA256 the password before passing it to createEncryptedHash() and
 * testEncryptedHash():
 * $hash = EncryptBcrypt::createEncryptedHash(hash('sha256', $pw));
 * if (EncryptBcrypt::testEncryptedHash(hash('sha256', $pw), $hash)) { ... }
 */
class EncryptBcrypt
{
	/**
	 * The derived 256 bit key for encrypting and decrypting hashes.
	 *
	 * @access private
	 * @var string
	 */
	private static $m_key;

	/**
	 * Context used in HKDF (an application specific string).
	 *
	 * @access private
	 * @var string
	 */
	private static $m_hkdfContext;

	/**
	 * Salt used in HKDF (picked at random and fixed forever).
	 *
	 * @access private
	 * @var string
	 */
	private static $m_hkdfSalt;


	// ***********************
	// **** Magic Methods ****
	// ***********************

	/**
	 * Can't create this class.
	 *
	 * @access private
	 */
	private function __construct() {}

	/**
	 * Can't clone this class.
	 *
	 * @access private
	 */
	private function __clone() {}

	/**
	 * Can't unserialize this class.
	 *
	 * @access private
	 */
	private function __wakeup() {}


	// ************************
	// **** Public Methods ****
	// ************************

	/**
	 * Creates an encrypted bcrypt hash.
	 *
	 * @access public
	 * @param string $pw    - The password to hash then encrypt.
	 * @param string $speed - The speed parameter for bcrypt a two decimal digits string and must be between 5 and 31 (default '09').
	 * @return string/bool  - On success returns the encrypted hash, otherwise false.
	 */
	public static function createEncryptedHash($pw, $speed = '09')
	{
		if (preg_match('/^[0-9]{2}$/', $speed) == 0 || intval($speed) < 5 || intval($speed) > 31)
		{
			$speed = '09';
		}
		$salt = strtr(base64_encode(self::hkdf_sha256(mcrypt_create_iv(55, MCRYPT_DEV_URANDOM), 128 + 8)), '+', '.'); // Never use MCRYPT_RAND
		$hash = crypt($pw, '$2y$' . $speed . '$' . $salt);
		if (!self::validBcryptHash($hash))
		{
			return false;
		}
		return self::encryptHash($hash, self::$m_key);
	}

	/**
	 * Creates an encrypted bcrypt hash from a bcrypt hash.
	 *
	 * @access public
	 * @param string $hash  - The hash to encrypt.
	 * @return string/bool  - On success returns the encrypted hash, otherwise false.
	 */
	public static function createEncryptedHashFromHash($hash)
	{
		if (!self::validBcryptHash($hash))
		{
			return false;
		}
		return self::encryptHash($hash, self::$m_key);
	}

	/**
	 * Derives and sets the encryption key from the source key using HKDF-SHA256 (HMAC-based Extract-and-Expand Key Derivation Function).
	 * THIS IS NOT FOR PASSWORDS!
	 *
	 * @access public
	 * @param string $sourceKey - High entropy random data that may not be uniformly distributed.
	 * @param string $context   - An application specific string such as "Example Project Name" or "example.com".
	 * @param string $salt      - Picked at random and fixed forever (default was the output of "mcrypt_create_iv(55, MCRYPT_DEV_URANDOM)").
	 */
	public static function setKey($sourceKey, $context = 'Example Project Name', $salt = "\xff\xe8\x71\x3e\xc1\x9b\xca\x2f\x06\x60\x2f\x54\xa4\x23\x3c\x3f\xb7\x3e\xef\x23\x56\xa6\xaa\x77\xe0\x2e\x04\x7c\xaa\xb0\x3c\x96\xb9\x32\xad\x83\xf0\x35\x9a\x0d\x8d\xe5\xa6\x2f\xee\x73\xb1\xca\x3d\x94\x82\xdc\x3f\x37\x41")
	{
		self::$m_hkdfContext = $context;
		self::$m_hkdfSalt    = $salt;
		self::$m_key         = self::hkdf_sha256($sourceKey, 256);
	}

	/**
	 * Checks if the password matches the hash.
	 *
	 * @access public
	 * @param string $pw            - The password to test.
	 * @param string $encryptedHash - The encrypted hash.
	 * @return bool                 - Returns true if the password is correct, otherwise false.
	 */
	public static function testEncryptedHash($pw, $encryptedHash)
	{
		$decryptedHash = self::decryptHash($encryptedHash, self::$m_key);
		if (is_int($decryptedHash))
		{
			return false;
		}
		return $decryptedHash == crypt($pw, $decryptedHash);
	}

	/**
	 * Checks if the hash is a valid bcrypt hash.
	 *
	 * @access public
	 * @param string $hash - The bcrypt hash.
	 * @return bool        - Returns true if the hash is valid, otherwise false.
	 */
	public static function validBcryptHash($hash)
	{
		return preg_match('/^\$2y\$[0-3][0-9]\$[\.\/0-9A-Za-z]{21}[\.Oeu][\.\/0-9A-Za-z]{30}[\.26CGKOSWaeimquy]$/', $hash) == 1 &&
			intval(substr($hash, 4, 2)) >= 5 && intval(substr($hash, 4, 2)) <= 31;
	}

	/**
	 * Checks if the encrypted hash is decryptable and a valid bcrypt hash.
	 *
	 * @access public
	 * @param string $encryptedHash - The encrypted hash.
	 * @return bool                 - Returns true if the encrypted hash is valid, otherwise false.
	 */
	public static function validEncryptedHash($encryptedHash)
	{
		return is_string(self::decryptHash($encryptedHash, self::$m_key));
	}

	/**
	 * Changes the encryption key for all the hashes in the database.
	 * **** Only use this when your database gets owned. ****
	 * Uses mysqli, if you need something else you'll need to change it.
	 * TODO - Code this in C++ to get like a 100x speed up.
	 *
	 * @access public
	 * @param string $dbHost           - Database host name/ip.
	 * @param string $dbUserName       - Database user name.
	 * @param string $dbPassword       - Database password.
	 * @param string $dbName           - Name of the database.
	 * @param string $userIdColumnName - The unique identifier for a user (this can be a id, uuid, user name, as long as it has an index on it and is unique).
	 * @param string $hashColumnName   - The encrypted hash column.
	 * @param string $userTableName    - The table that contains the encrypted hashes.
	 * @param string $resume           - In the event that this script crashes part way through set this to true (default false).
	 * @return int                     - On success returns 0, otherwise 1.
	 */
	public static function changeDatabaseEncryptionKey($dbHost, $dbUserName, $dbPassword, $dbName, $userIdColumnName, $hashColumnName, $userTableName, $resume = false)
	{
		if (!$resume)
		{
			if (file_exists('set_server_key-old.php'))
			{
				echo "set_server_key-old.php already exists. This means that this is already running. If it stopped or crashed you can \"resume\" by setting \$resume to true.<br>\n";
				return 1;
			}
			if (!rename('set_server_key.php', 'set_server_key-old.php'))
			{
				echo "Can't rename set_server_key.php<br>\n";
				return 1;
			}
			file_put_contents('set_server_key.php', '<?php EncryptBcrypt::setKey(base64_decode(\'' . base64_encode(mcrypt_create_iv(55, MCRYPT_DEV_URANDOM)) . '\'));');
		}
		require 'set_server_key-old.php';
		$oldServerKey = self::$m_key;
		require 'set_server_key.php';
		$newServerKey = self::$m_key;

		$mysqliIn = new mysqli($dbHost, $dbUserName, $dbPassword, $dbName);
		if ($mysqliIn->connect_errno)
		{
			echo 'Connect failed: ' . $mysqliIn->connect_error . "<br>\n";
			return 1;
		}
		$mysqliOut = new mysqli($dbHost, $dbUserName, $dbPassword, $dbName);
		if ($mysqliOut->connect_errno)
		{
			echo 'Connect failed: ' . $mysqliOut->connect_error . "<br>\n";
			return 1;
		}

		$result = $mysqliIn->query("SELECT `$userIdColumnName`, `$hashColumnName` FROM `$userTableName`", MYSQLI_USE_RESULT);
		if (!$result)
		{
			echo "MySQL error: " . $mysqliIn->error . "<br>\n";
			return 1;
		}
		$row = $result->fetch_row();
		if ($resume)
		{
			do
			{
				$hash = self::changeHashEncryptionKey($row[1], $oldServerKey, $newServerKey);
			} while ($hash === 2 && $row = $result->fetch_row());
			if (!$row)
			{
				echo "Interesting did you forget to securely delete set_server_key-old.php the last time your database got owned?<br>\n" .
					"Either that or you tried to rerun this script or this crashed after it finished updated the database?<br>\n" .
					"If this is done, please securely delete set_server_key-old.php (you may want to back this up on an offline medium such as a flash drive just incase something went wrong)<br>\n";
				return 1;
			}
		}
		$error = false;
		do
		{
			$hash = self::changeHashEncryptionKey($row[1], $oldServerKey, $newServerKey);
			if ($hash === 2)
			{
				if (!$error)
				{
					echo "This maybe bad either a corrupt database or failed to set \$resume to true and removed set_server_key-old.php which means that you just lost a lot of hashes.<br>\n";
				}
				$error = true;
				echo 'Error: Bad hash for `$userIdColumnName` = "' . $row[0] . "\"<br>\n";
			}
			else
			{
				$hash = $mysqliOut->real_escape_string($hash);
				if ($hash !== 1 && !$mysqliOut->query("UPDATE `$userTableName` SET `$hashColumnName`='$hash' WHERE `$userIdColumnName`='" . $row[0] . "'"))
				{
					echo 'Error: ' . $mysqliOut->error . "<br>\n";
					return 1;
				}
			}
		} while ($row = $result->fetch_row());
		$result->close();
		$mysqliOut->close();
		$mysqliIn->close();
		echo "Done, if you want to you can test the encryption key with testEncryptionKey() (I suggest you do). After you have tested the encryption key, please securely delete set_server_key-old.php (you may want to back this up on an offline medium such as a flash drive just incase something went wrong)<br>\n";
		return 0;
	}

	/**
	 * Test the encryption key for all the hashes in the database.
	 * Uses mysqli, if you need something else you'll need to change it.
	 * TODO - Code this in C++ to get like a 100x speed up
	 *
	 * @access public
	 * @param string $dbHost           - Database host name/ip.
	 * @param string $dbUserName       - Database user name.
	 * @param string $dbPassword       - Database password.
	 * @param string $dbName           - Name of the database.
	 * @param string $userIdColumnName - The unique identifier for a user (this can be a id, uuid, user name, as long as it has an index on it and is unique).
	 * @param string $hashColumnName   - The encrypted hash column.
	 * @param string $userTableName    - The table that contains the encrypted hashes.
	 * @param string $outputBadIds     - Output user id's that have bad hashes (default false).
	 * @return int                     - On success returns number of failed hashes, otherwise -1.
	 */
	public static function testDatabaseEncryptionKey($dbHost, $dbUserName, $dbPassword, $dbName, $userIdColumnName, $hashColumnName, $userTableName, $outputBadIds = false)
	{
		$mysqliIn = new mysqli($dbHost, $dbUserName, $dbPassword, $dbName);
		if ($mysqliIn->connect_errno)
		{
			echo 'Connect failed: ' . $mysqliIn->connect_error . "<br>\n";
			return -1;
		}

		$result = $mysqliIn->query("SELECT `$userIdColumnName`, `$hashColumnName` FROM `$userTableName`", MYSQLI_USE_RESULT);
		if (!$result)
		{
			echo "MySQL error: " . $mysqliIn->error . "<br>\n";
			return -1;
		}
		$error = false;
		$badCount = 0;
		while ($row = $result->fetch_row())
		{
			if (self::decryptHash($row[1], self::$m_key) === 2)
			{
				if ($outputBadIds)
				{
					if (!$error)
					{
						echo "This is bad either a corrupt database or hashes were encrypted with a different key.<br>\n";
					}
					$error = true;
					echo 'Error: Bad hash for `$userIdColumnName` = "' . $row[0] . "\"<br>\n";
				}
				$badCount++;
			}
		}
		$result->close();
		$mysqliIn->close();
		return $badCount;
	}


	// *************************
	// **** Private Methods ****
	// *************************

	/**
	 * Changes the encryption key for the given hash.
	 *
	 * @access private
	 * @param string $encryptedHash - The encrypted hash.
	 * @param string $oldServerKey  - The old key to decrypt the hash.
	 * @param string $newServerKey  - The new key to encrypt the hash.
	 * @return string/int           - On success returns the newly encrypted hash.
	 *                                1 - if the encrypted hash is not a valid encrypted hash.
	 *                                2 - if the encryption key is not correct or the decrypted hash is not a valid bcrypt hash.
	 */
	private static function changeHashEncryptionKey($encryptedHash, $oldServerKey, $newServerKey)
	{
		$decryptedHash = self::decryptHash($encryptedHash, $oldServerKey);
		if (is_int($decryptedHash))
		{
			return $decryptedHash;
		}
		return self::encryptHash($decryptedHash, $newServerKey);
	}

	/**
	 * Decrypts a hash.
	 *
	 * @access private
	 * @param string $encryptedHash - The hash to decrypt.
	 * @param string $key           - The key to decrypt the hash.
	 * @return string/int           - On success returns the decrypted hash.
	 *                                1 - if the encrypted hash is not a valid encrypted hash.
	 *                                2 - if the encryption key is not correct or the decrypted hash is not a valid bcrypt hash.
	 */
	private static function decryptHash($encryptedHash, $key)
	{
		if (substr($encryptedHash, 0, 14) != '$e$aes256-cbc$')
		{
			return 1;
		}
		$encryptedData = base64_decode(substr($encryptedHash, 14));
		$iv = substr($encryptedData, 0, 16);
		$ct = substr($encryptedData, 16);
		$decryptedHash = mcrypt_cbc(MCRYPT_RIJNDAEL_128, $key, $ct, MCRYPT_DECRYPT, $iv); // AES 256 CBC

		// Fix for the PHP padding bug
		//   Seriously, mcrypt_cbc should use the f-ing standard or the other f-ing standard:
		//     $pt = $pt . str_repeat(chr(BLOCK_SIZE - strlen($pt) % BLOCK_SIZE), BLOCK_SIZE - strlen($pt) % BLOCK_SIZE);
		//     $pt = $pt . "\x80" . str_repeat("\0", BLOCK_SIZE - 1 - strlen($pt) % BLOCK_SIZE);
		$decryptedHash = rtrim($decryptedHash, "\0");

		if (!self::validBcryptHash($decryptedHash))
		{
			return 2;
		}
		return $decryptedHash;
	}

	/**
	 * Encrypts a hash.
	 *
	 * @access private
	 * @param string $hash - The hash to encrypt.
	 * @param string $key  - The key to encrypt the hash.
	 * @return string      - Returns the encrypted hash.
	 */
	private static function encryptHash($hash, $key)
	{
		$iv = self::hkdf_sha256(mcrypt_create_iv(55, MCRYPT_DEV_URANDOM), 128); // Never use MCRYPT_RAND
		return '$e$aes256-cbc$' . base64_encode($iv . mcrypt_cbc(MCRYPT_RIJNDAEL_128, $key, $hash, MCRYPT_ENCRYPT, $iv)); // AES 256 CBC
	}

	/**
	 * Key derivation function (THIS IS NOT FOR PASSWORDS).
	 *
	 * @access private
	 * @param string $sourceKey  - High entropy random data that may not be uniformly distributed.
	 * @param int    $keyBitSize - Key size in bits (will be rounded up to the next multiple of 8).
	 * @return string            - The derived key.
	 */
	private static function hkdf_sha256($sourceKey, $keyBitSize)
	{
		$keySize = (intval($keyBitSize) + 7) >> 3;
		$extractedKey = hash_hmac('sha256', self::$m_hkdfSalt, $sourceKey, true);
		$hmacSize = strlen($extractedKey);
		$key = str_repeat("\0", $hmacSize);
		for ($i = 1; strlen($key) - $hmacSize < $keySize; $i++)
		{
			$key .= hash_hmac('sha256', substr($key, -$hmacSize) . self::$m_hkdfContext . chr($i), $extractedKey, true);
		}
		return substr($key, $hmacSize, $keySize);
	}
}

// * This is optional but you need to have some file somewhere that has the key
//   so that you can pass it into setKey().
// * You will need to modify changeEncryptionKey() if you do not do it this way.
// * If you have multiple web servers then this will not work for you since each
//   server will generate their own key.
if (!file_exists('set_server_key.php'))
{
	file_put_contents('set_server_key.php', '<?php EncryptBcrypt::setKey(base64_decode(\'' . base64_encode(mcrypt_create_iv(55, MCRYPT_DEV_URANDOM)) . '\'));');
}
require_once 'set_server_key.php';

?>